function [L2dist,ddh]=LSDD(X1,X2,T)
%
% Least-Squares Density-Difference Estimation
%
% Estimating p1(x)-p2(x) from samples {x1_i}_{i=1}^{n1} and {x2_j}_{j=1}^{n2}
% drawn i.i.d. from p1(x) and p2(x), respectively.
%
% Usage:
%       [L2dist,ddh]=LSDD(X1,X2)
%
% Input:
%    X1: d by n1 training sample matrix
%    X2: d by n2 training sample matrix
%    T:  d by n2 test sample matrix (OPTIONAL, default: T=[X1 X2])
%
% Output:
%    L2dist: estimate of L2-distance obtained from X1 and X2
%    ddh:    estimates of p1(x)-p2(x) at T (OPTIONAL)
%
% (c) Masashi Sugiyama, Department of Compter Science, Tokyo Institute of Technology, Japan.
%     sugi@cs.titech.ac.jp,     http://sugiyama-www.cs.titech.ac.jp/~sugi/software/LSDD/

[d,n1]=size(X1);
[dummy,n2]=size(X2);
X=[X1 X2];
clear X1 X2
n=n1+n2;
b=min(300,n); % Number of kernel bases

center_index_tmp=randperm(n);
center_index=center_index_tmp(1:b);
C=X(:,center_index); 
XXsum=sum(X.^2,1);
XC_dist2=repmat(XXsum,[b 1])+repmat(XXsum(1,center_index)',[1 n])-2*C'*X;

if nargout>=2 & nargin>=3
  TC_dist2=repmat(sum(T.^2,1),[b 1])+repmat(XXsum(1,center_index)',[1 size(T,2)])-2*C'*T;
end
clear XXsum X C T
X1C_dist2=XC_dist2(:,1:n1);
X2C_dist2=XC_dist2(:,n1+1:end);
CC_dist2=XC_dist2(:,center_index);
clear XC_dist2

fold=5;
cv_fold=[1:fold];
cv_split1=floor([0:n1-1]*fold./n1)+1;
cv_split2=floor([0:n2-1]*fold./n2)+1;
cv_index1=cv_split1(randperm(n1));
cv_index2=cv_split2(randperm(n2));
for k=cv_fold
  n1_cv(k)=sum(cv_index1==k);
  n2_cv(k)=sum(cv_index2==k);
end % k

sigma_list=[1/4 1/2 3/4 1 1.2 1.5 2 3 5]; % Candidates of Gaussian width
lambda_list=logspace(-3,1,9); % Candidates of regularization parameter


score_cv=zeros(length(sigma_list),length(lambda_list),length(cv_fold));
for sigma_index=1:length(sigma_list)
  sigma=sigma_list(sigma_index);
  H=(sqrt(pi)*sigma)^d*exp(-CC_dist2/(4*sigma^2));
  for k=cv_fold
    h1_cv(:,k)=sum(exp(-X1C_dist2(:,cv_index1==k)/(2*sigma^2)),2);
    h2_cv(:,k)=sum(exp(-X2C_dist2(:,cv_index2==k)/(2*sigma^2)),2);
  end % k
  for k=cv_fold
    htrain=sum(h1_cv(:,cv_fold~=k),2)/sum(n1_cv(cv_fold~=k))...
           -sum(h2_cv(:,cv_fold~=k),2)/sum(n2_cv(cv_fold~=k));
    htest=sum(h1_cv(:,cv_fold==k),2)/sum(n1_cv(cv_fold==k))...
          -sum(h2_cv(:,cv_fold==k),2)/sum(n2_cv(cv_fold==k));
    for lambda_index=1:length(lambda_list)
      lambda=lambda_list(lambda_index);
      thetah=mylinsolve(H+lambda*eye(b),htrain);
      score_cv(sigma_index,lambda_index,k)=thetah'*H*thetah-2*thetah'*htest;
    end % lambda
  end % k
end % sigma

[score_cv_tmp,lambda_index]=min(mean(score_cv,3),[],2);
[score,sigma_index]=min(score_cv_tmp);
lambda=lambda_list(lambda_index(sigma_index));
sigma=sigma_list(sigma_index);
%disp(sprintf('sigma = %g, lambda = %g',sigma,lambda))

H=(sqrt(pi)*sigma)^d*exp(-CC_dist2/(4*sigma^2));
h=mean(exp(-X1C_dist2/(2*sigma^2)),2)-mean(exp(-X2C_dist2/(2*sigma^2)),2);
thetah=mylinsolve(H+lambda*eye(b),h);
L2dist=2*thetah'*h-thetah'*H*thetah;

if nargout>=2
  if nargin<3
    ddh=thetah'*[exp(-[X1C_dist2 X2C_dist2]/(2*sigma^2))];
  else
    ddh=thetah'*[exp(-TC_dist2/(2*sigma^2))];
  end
end
